/**********************************************************************
Copyright (c) 2009 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.odf;

import java.sql.Timestamp;
import java.util.Date;

import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusObjectNotFoundException;
import org.datanucleus.exceptions.NucleusUserException;
import org.datanucleus.identity.OID;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.IdentityType;
import org.datanucleus.metadata.VersionMetaData;
import org.datanucleus.metadata.VersionStrategy;
import org.datanucleus.store.AbstractPersistenceHandler;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.StoreManager;
import org.datanucleus.store.connection.ManagedConnection;
import org.datanucleus.store.fieldmanager.DeleteFieldManager;
import org.datanucleus.store.odf.fieldmanager.FetchFieldManager;
import org.datanucleus.store.odf.fieldmanager.InsertFieldManager;
import org.datanucleus.store.odf.fieldmanager.UpdateFieldManager;
import org.datanucleus.util.Localiser;
import org.datanucleus.util.NucleusLogger;
import org.datanucleus.util.StringUtils;
import org.openoffice.odf.doc.OdfDocument;
import org.openoffice.odf.doc.OdfFileDom;
import org.openoffice.odf.doc.OdfSpreadsheetDocument;
import org.openoffice.odf.doc.element.table.OdfTable;
import org.openoffice.odf.doc.element.table.OdfTableCell;
import org.openoffice.odf.doc.element.table.OdfTableRow;
import org.openoffice.odf.dom.OdfNamespace;
import org.openoffice.odf.dom.type.office.OdfValueType;
import org.w3c.dom.Element;

/**
 * Persistence Handler for Open Document Format (ODF) datastores.
 * Handles the insert/update/delete/fetch/locate operations by using ODF Toolkit.
 * <h3>Field -> Cell mapping</h3>
 * A field is mapped to a cell. The field metadata can define a cell number (starting at 0).
 * Specifying the cell number means that the user takes responsibility for the cell numbers
 * being consistent. The default cell numbering is alphabetical start in the root class, and working
 * down the inheritance tree to the actual instance class.
 */
public class ODFPersistenceHandler extends AbstractPersistenceHandler
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER = Localiser.getInstance(
        "org.datanucleus.store.odf.Localisation", ODFStoreManager.class.getClassLoader());

    protected final ODFStoreManager storeMgr;

    /**
     * Constructor.
     * @param storeMgr Manager for the datastore
     */
    public ODFPersistenceHandler(StoreManager storeMgr)
    {
        this.storeMgr = (ODFStoreManager)storeMgr;
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.StorePersistenceHandler#close()
     */
    public void close()
    {
        // Nothing to do since we maintain no resources
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.StorePersistenceHandler#insertObject(org.datanucleus.StateManager)
     */
    public void insertObject(ObjectProvider sm)
    {
        // Check if read-only so update not permitted
        storeMgr.assertReadOnlyForUpdateOfObject(sm);

        AbstractClassMetaData cmd = sm.getClassMetaData();
        if (cmd.getIdentityType() == IdentityType.APPLICATION || cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            // Enforce uniqueness of datastore rows
            try
            {
                locateObject(sm);
                throw new NucleusUserException(LOCALISER.msg("ODF.Insert.ObjectWithIdAlreadyExists",
                    sm.toPrintableID(), sm.getInternalObjectId()));
            }
            catch (NucleusObjectNotFoundException onfe)
            {
                // Do nothing since object with this id doesn't exist
            }
        }

        ExecutionContext ec = sm.getExecutionContext();
        ManagedConnection mconn = storeMgr.getConnection(ec);
        try
        {
            long startTime = System.currentTimeMillis();
            if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("ODF.Insert.Start", 
                    sm.toPrintableID(), sm.getInternalObjectId()));
            }

            OdfSpreadsheetDocument doc = (OdfSpreadsheetDocument)mconn.getConnection();
            OdfFileDom contentDoc = doc.getContentDom();

            // Find the sheet (table) appropriate for storing objects of this class
            String sheetName = ODFUtils.getSheetNameForClass(cmd);
            OdfTable table = ODFUtils.getODFTableForClass(cmd, contentDoc);
            if (table == null)
            {
                // Table doesn't exist yet so create and add to spreadsheet
                table = new OdfTable(contentDoc);
                table.setName(sheetName);
                Element officeSpreadsheet = (Element)contentDoc.getElementsByTagNameNS(
                    OdfNamespace.OFFICE.getUri(), "spreadsheet").item(0);
                officeSpreadsheet.appendChild(table);

                if (cmd.hasExtension("include-column-headers") &&
                    cmd.getValueForExtension("include-column-headers").equalsIgnoreCase("true"))
                {
                    // Add extra row with headers
                    addHeaderRowForTable(contentDoc, cmd, table);
                }

                if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
                {
                    NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("ODF.Insert.SheetCreated",
                        sm.toPrintableID(), sheetName));
                }
            }

            // Add a new row to this table for this object
            OdfTableRow row = new OdfTableRow(contentDoc);
            table.appendChild(row);

            // Add cells for the fields to this row
            InsertFieldManager insertFM = new InsertFieldManager(sm, contentDoc, row);
            sm.provideFields(cmd.getAllMemberPositions(), insertFM);

            if (cmd.getIdentityType() == IdentityType.DATASTORE)
            {
                int colIndex = ODFUtils.getColumnIndexForFieldOfClass(cmd, -1);
                OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
                Object idKey = ((OID)sm.getInternalObjectId()).getKeyValue();
                if (idKey instanceof String)
                {
                    cell.setValueType(OdfValueType.STRING);
                    cell.setStringValue((String)idKey);
                }
                else
                {
                    long idValue = ((Long)((OID)sm.getInternalObjectId()).getKeyValue()).longValue();
                    cell.setValueType(OdfValueType.FLOAT);
                    cell.setValue(new Double(idValue));
                }
                cell.setStyleName("DN_PK");
            }
            if (cmd.hasVersionStrategy())
            {
                int colIndex = ODFUtils.getColumnIndexForFieldOfClass(cmd, -2);
                OdfTableCell cell = (OdfTableCell)row.getChildNodes().item(colIndex);
                if (cmd.getVersionMetaData().getVersionStrategy() == VersionStrategy.VERSION_NUMBER)
                {
                    long versionNumber = 1;
                    sm.setTransactionalVersion(Long.valueOf(versionNumber));
                    if (NucleusLogger.DATASTORE.isDebugEnabled())
                    {
                        NucleusLogger.DATASTORE.debug(LOCALISER.msg("ODF.Insert.ObjectPersistedWithVersion",
                            sm.toPrintableID(), sm.getInternalObjectId(), "" + versionNumber));
                    }
                    cell.setValueType(OdfValueType.FLOAT);
                    cell.setValue(new Double(versionNumber));
                }
                else if (cmd.getVersionMetaData().getVersionStrategy() == VersionStrategy.DATE_TIME)
                {
                    Date date = new Date();
                    Timestamp ts = new Timestamp(date.getTime());
                    sm.setTransactionalVersion(ts);
                    if (NucleusLogger.DATASTORE.isDebugEnabled())
                    {
                        NucleusLogger.DATASTORE.debug(LOCALISER.msg("ODF.Insert.ObjectPersistedWithVersion",
                            sm.toPrintableID(), sm.getInternalObjectId(), "" + ts));
                    }
                    cell.setValueType(OdfValueType.FLOAT);
                    cell.setValue(new Double(ts.getTime()));
                }
            }

            if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("ODF.ExecutionTime", 
                    (System.currentTimeMillis() - startTime)));
            }
            if (storeMgr.getRuntimeManager() != null)
            {
                storeMgr.getRuntimeManager().incrementInsertCount();
            }
            if (NucleusLogger.DATASTORE.isDebugEnabled())
            {
                NucleusLogger.DATASTORE.debug(LOCALISER.msg("ODF.Insert.ObjectPersisted",
                    sm.toPrintableID(), sm.getInternalObjectId()));
            }

            // Use SCO wrappers from this point
            sm.replaceAllLoadedSCOFieldsWithWrappers();
        }
        catch (Exception e)
        {
            throw new NucleusDataStoreException("Error inserting object", e);
        }
        finally
        {
            mconn.release();
        }
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.StorePersistenceHandler#updateObject(org.datanucleus.StateManager, int[])
     */
    public void updateObject(ObjectProvider sm, int[] fieldNumbers)
    {
        // Check if read-only so update not permitted
        storeMgr.assertReadOnlyForUpdateOfObject(sm);

        ManagedConnection mconn = storeMgr.getConnection(sm.getExecutionContext());
        try
        {
            OdfDocument doc = (OdfDocument)mconn.getConnection();
            OdfFileDom contentDoc = doc.getContentDom();

            // TODO Add optimistic checks
            Object currentVersion = sm.getTransactionalVersion();
            Object nextVersion = null;
            AbstractClassMetaData acmd = sm.getClassMetaData();
            if (acmd.hasVersionStrategy())
            {
                // Version object so calculate version to store with
                VersionMetaData vermd = acmd.getVersionMetaData();
                if (acmd.getVersionMetaData().getFieldName() != null)
                {
                    // Version field
                    AbstractMemberMetaData verfmd = acmd.getMetaDataForMember(vermd.getFieldName());
                    if (currentVersion instanceof Integer)
                    {
                        // Cater for Integer-based versions TODO Generalise this
                        currentVersion = Long.valueOf(((Integer)currentVersion).longValue());
                    }

                    nextVersion = acmd.getVersionMetaData().getNextVersion(currentVersion);
                    if (verfmd.getType() == Integer.class || verfmd.getType() == int.class)
                    {
                        // Cater for Integer-based versions TODO Generalise this
                        nextVersion = Integer.valueOf(((Long)nextVersion).intValue());
                    }
                    sm.replaceField(verfmd.getAbsoluteFieldNumber(), nextVersion);
                }
                else
                {
                    // Surrogate version column
                    nextVersion = vermd.getNextVersion(currentVersion);
                }
                sm.setTransactionalVersion(nextVersion);
            }

            long startTime = System.currentTimeMillis();
            if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
            {
                StringBuffer fieldStr = new StringBuffer();
                for (int i=0;i<fieldNumbers.length;i++)
                {
                    if (i > 0)
                    {
                        fieldStr.append(",");
                    }
                    fieldStr.append(acmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumbers[i]).getName());
                }
                NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("ODF.Update.Start", 
                    sm.toPrintableID(), sm.getInternalObjectId(), fieldStr.toString()));
            }

            // Update the row in the worksheet
            OdfTableRow row = ODFUtils.getTableRowForObjectInSheet(sm, contentDoc);
            if (row == null)
            {
                String sheetName = ODFUtils.getSheetNameForClass(sm.getClassMetaData());
                throw new NucleusDataStoreException(LOCALISER.msg("ODF.RowNotFoundForSheetForWorkbook",
                    sheetName, StringUtils.toJVMIDString(sm.getInternalObjectId())));
            }
            sm.provideFields(fieldNumbers, new UpdateFieldManager(sm, row));

            if (acmd.hasVersionStrategy())
            {
                // Versioned object so set version cell in spreadsheet
                int verCellIndex = ODFUtils.getColumnIndexForFieldOfClass(acmd, -2);
                OdfTableCell verCell = (OdfTableCell)row.getChildNodes().item(verCellIndex);
                if (acmd.getVersionMetaData().getVersionStrategy() == VersionStrategy.VERSION_NUMBER)
                {
                    sm.setTransactionalVersion(nextVersion);
                    if (NucleusLogger.DATASTORE.isDebugEnabled())
                    {
                        NucleusLogger.DATASTORE.debug(LOCALISER.msg("ODF.Insert.ObjectPersistedWithVersion",
                            sm.toPrintableID(), sm.getInternalObjectId(), 
                            "" + nextVersion));
                    }
                    verCell.setValue(new Double((Long)nextVersion));
                }
                else if (acmd.getVersionMetaData().getVersionStrategy() == VersionStrategy.DATE_TIME)
                {
                    sm.setTransactionalVersion(nextVersion);
                    if (NucleusLogger.DATASTORE.isDebugEnabled())
                    {
                        NucleusLogger.DATASTORE.debug(LOCALISER.msg("ODF.Insert.ObjectPersistedWithVersion",
                            sm.toPrintableID(), sm.getInternalObjectId(), "" + nextVersion));
                    }
                    Timestamp ts = (Timestamp)nextVersion;
                    Date date = new Date();
                    date.setTime(ts.getTime());
                    verCell.setValue(new Double(ts.getTime()));
                }
            }

            if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("ODF.ExecutionTime", 
                    (System.currentTimeMillis() - startTime)));
            }
            if (storeMgr.getRuntimeManager() != null)
            {
                storeMgr.getRuntimeManager().incrementUpdateCount();
            }
        }
        catch (Exception e)
        {
            throw new NucleusDataStoreException("Exception updating object", e);
        }
        finally
        {
            mconn.release();
        }
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.StorePersistenceHandler#deleteObject(org.datanucleus.StateManager)
     */
    public void deleteObject(ObjectProvider sm)
    {
        // Check if read-only so update not permitted
        storeMgr.assertReadOnlyForUpdateOfObject(sm);

        ManagedConnection mconn = storeMgr.getConnection(sm.getExecutionContext());
        try
        {
            OdfDocument doc = (OdfDocument)mconn.getConnection();
            OdfFileDom contentDoc = doc.getContentDom();

            // TODO Add optimistic checks

            // Delete all reachable PC objects (due to dependent-field)
            sm.loadUnloadedFields();
            sm.provideFields(sm.getClassMetaData().getAllMemberPositions(), new DeleteFieldManager(sm));

            // Delete this object
            long startTime = System.currentTimeMillis();
            if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("ODF.Delete.Start", 
                    sm.toPrintableID(), sm.getInternalObjectId()));
            }

            OdfTableRow row = ODFUtils.getTableRowForObjectInSheet(sm, contentDoc);
            if (row == null)
            {
                throw new NucleusObjectNotFoundException("object not found", sm.getObject());
            }
            else
            {
                // Remove the row node
                OdfTable table = ODFUtils.getODFTableForClass(sm.getClassMetaData(), contentDoc);
                table.removeChild(row);
            }

            if (NucleusLogger.DATASTORE_PERSIST.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_PERSIST.debug(LOCALISER.msg("ODF.ExecutionTime", 
                    (System.currentTimeMillis() - startTime)));
            }
            if (storeMgr.getRuntimeManager() != null)
            {
                storeMgr.getRuntimeManager().incrementDeleteCount();
            }
        }
        catch (Exception e)
        {
            throw new NucleusDataStoreException("Exception deleting object", e);
        }
        finally
        {
            mconn.release();
        }
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.StorePersistenceHandler#fetchObject(org.datanucleus.StateManager, int[])
     */
    public void fetchObject(ObjectProvider sm, int[] fieldNumbers)
    {
        AbstractClassMetaData acmd = sm.getClassMetaData();
        if (NucleusLogger.PERSISTENCE.isDebugEnabled())
        {
            // Debug information about what we are retrieving
            StringBuffer str = new StringBuffer("Fetching object \"");
            str.append(sm.toPrintableID()).append("\" (id=");
            str.append(sm.getExecutionContext().getApiAdapter().getObjectId(sm)).append(")").append(" fields [");
            for (int i=0;i<fieldNumbers.length;i++)
            {
                if (i > 0)
                {
                    str.append(",");
                }
                str.append(acmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumbers[i]).getName());
            }
            str.append("]");
            NucleusLogger.PERSISTENCE.debug(str);
        }

        ManagedConnection mconn = storeMgr.getConnection(sm.getExecutionContext());
        boolean notFound = false;
        try
        {
            OdfDocument doc = (OdfDocument)mconn.getConnection();
            OdfFileDom contentDoc = doc.getContentDom();

            long startTime = System.currentTimeMillis();
            if (NucleusLogger.DATASTORE_RETRIEVE.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_RETRIEVE.debug(LOCALISER.msg("ODF.Fetch.Start", 
                    sm.toPrintableID(), sm.getInternalObjectId()));
            }

            OdfTableRow row = ODFUtils.getTableRowForObjectInSheet(sm, contentDoc);
            if (row == null)
            {
                notFound = true;
            }
            else
            {
                sm.replaceFields(fieldNumbers, new FetchFieldManager(sm, row));

                if (NucleusLogger.DATASTORE_RETRIEVE.isDebugEnabled())
                {
                    NucleusLogger.DATASTORE_RETRIEVE.debug(LOCALISER.msg("ODF.ExecutionTime", 
                        (System.currentTimeMillis() - startTime)));
                }
                if (storeMgr.getRuntimeManager() != null)
                {
                    storeMgr.getRuntimeManager().incrementFetchCount();
                }
                // TODO Version retrieval
            }
        }
        catch (Exception e)
        {
            throw new NucleusDataStoreException("Exception fetching object", e);
        }
        finally
        {
            mconn.release();
        }

        if (notFound)
        {
            throw new NucleusObjectNotFoundException("object not found", sm.getObject());
        }
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.StorePersistenceHandler#findObject(org.datanucleus.ObjectManager, java.lang.Object)
     */
    public Object findObject(ExecutionContext ec, Object id)
    {
        return null;
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.StorePersistenceHandler#locateObject(org.datanucleus.StateManager)
     */
    public void locateObject(ObjectProvider sm)
    {
        final AbstractClassMetaData acmd = sm.getClassMetaData();
        if (acmd.getIdentityType() == IdentityType.APPLICATION ||
            acmd.getIdentityType() == IdentityType.DATASTORE)
        {
            ManagedConnection mconn = storeMgr.getConnection(sm.getExecutionContext());
            try
            {
                OdfDocument doc = (OdfDocument)mconn.getConnection();
                OdfFileDom contentDoc = doc.getContentDom();
                OdfTableRow row = ODFUtils.getTableRowForObjectInSheet(sm, contentDoc);
                if (row != null)
                {
                    return;
                }
            }
            catch (Exception e)
            {
                e.printStackTrace();
                throw new NucleusDataStoreException("Error when trying to find object with id=" + sm.getInternalObjectId());
            }
            finally
            {
                mconn.release();
            }
        }

        throw new NucleusObjectNotFoundException("Object not found", sm.getExternalObjectId());
    }

    /**
     * Method to add a row to the specified table providing column headers using the metadata for the
     * class that it is for. Uses the column metadata name if provided otherwise the field name
     * @param contentDoc DOM for the content
     * @param cmd Metadata for the class stored in the table
     * @param table The table that this adds a row to
     */
    protected void addHeaderRowForTable(OdfFileDom contentDoc, AbstractClassMetaData cmd, OdfTable table)
    {
        // Add extra row with headers
        OdfTableRow row = new OdfTableRow(contentDoc);
        table.appendChild(row);
        int numFields = cmd.getAllMemberPositions().length;
        int numCols = numFields;
        if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            numCols++;
        }
        if (cmd.hasVersionStrategy())
        {
            numCols++;
        }
        OdfTableCell[] cells = new OdfTableCell[numCols];
        for (int i=0;i<numFields;i++)
        {
            int index = ODFUtils.getColumnIndexForFieldOfClass(cmd, i);
            AbstractMemberMetaData mmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(i);
            cells[index] = new OdfTableCell(contentDoc);
            cells[index].setValueType(OdfValueType.STRING);
            String colName = mmd.getName();
            ColumnMetaData[] colmds = mmd.getColumnMetaData();
            if (colmds != null && colmds.length > 0 && colmds[0].getName() != null)
            {
                colName = colmds[0].getName();
            }
            cells[index].setStringValue(colName);
        }
        if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            int index = ODFUtils.getColumnIndexForFieldOfClass(cmd, -1);
            cells[index] = new OdfTableCell(contentDoc);
            cells[index].setValueType(OdfValueType.STRING);
            String idName = "ID";
            ColumnMetaData[] colmds = cmd.getIdentityMetaData().getColumnMetaData();
            if (colmds != null && colmds.length > 0 && colmds[0].getName() != null)
            {
                idName = colmds[0].getName();
            }
            cells[index].setStringValue(idName);
        }
        if (cmd.hasVersionStrategy())
        {
            int index = ODFUtils.getColumnIndexForFieldOfClass(cmd, -1);
            cells[index] = new OdfTableCell(contentDoc);
            cells[index].setValueType(OdfValueType.STRING);
            String idName = "VERSION";
            ColumnMetaData[] colmds = cmd.getVersionMetaData().getColumnMetaData();
            if (colmds != null && colmds.length > 0 && colmds[0].getName() != null)
            {
                idName = colmds[0].getName();
            }
            cells[index].setStringValue(idName);
        }

        for (int i=0;i<cells.length;i++)
        {
            row.appendCell(cells[i]);
        }
        row.setStyleName("DN_Headers");
    }
}